#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright(C) 2006 INL
Written by Damien Boucard <damien.boucard AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
---
auto_daemon.py is generic class to manage daemon script.

Features available with command line :
  * Choosing a verbosity level
  * Setting the logfile location
  * Setting the pidfile location
  * Choosing not to detach daemon in background (for debugging purpose)

This file can be executed to see how to use it.
"""

from optparse import OptionParser, Values
import os, os.path, sys, time, logging, signal

__revision__ = "0.1"

class AutoDaemon:
    # Constant module attributes
    APP_NAME = None
    LEVELS = (logging.NOTSET, logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL)
    DEFAULT_OPTIONS = { 'verbosity': 0,
                        'daemonize': True,
                        'log': "/var/log/%(app_name)s.log",
                        'pidfile': "/var/run/%(app_name)s.pid" }
    
    def __init__(self, application_name):
        self.APP_NAME = application_name
        self.APP_ABSPATH = os.path.abspath(sys.argv[0])
        if hasattr(self, 'options'):
            raise AttributeError, "'options' attribute already defined."
        self.DEFAULT_OPTIONS = self.DEFAULT_OPTIONS.copy()
        self.DEFAULT_OPTIONS['log'] = self.DEFAULT_OPTIONS['log'] % {'app_name': self.APP_NAME}
        self.DEFAULT_OPTIONS['pidfile'] = self.DEFAULT_OPTIONS['pidfile'] % {'app_name': self.APP_NAME}
        self.options = Values(self.DEFAULT_OPTIONS)
        self.arguments = ()
        
    def setOptions(self, **kwargs):
        for kw in kwargs:
            if kw in self.DEFAULT_OPTIONS:
                setattr(self.options, kw, kwargs[kw])
            else:
                raise TypeError, "setOptions() got an unexpected keyword argument '%s'" %(kw)
        
    def parseOptions(self):
        # Use the method if command line script
        parser = OptionParser(usage="%prog [options]")
    
        # Base Options
        parser.add_option("-v", "--verbosity",
            help="Verbosity mode [0..%s]" %(len(self.LEVELS)-1),
            action="store", type="int", metavar="LEVEL",
            default=self.DEFAULT_OPTIONS['verbosity'])
        if self.DEFAULT_OPTIONS['daemonize']:
            parser.add_option("-d", "--not-daemonize",
                help="Don't launch the program in background",
                action="store_true",
                default=not self.DEFAULT_OPTIONS['daemonize'])
        else:
            parser.add_option("-d", "--daemonize",
                help="Lunch the program in background",
                action="store_true",
                default=self.DEFAULT_OPTIONS['daemonize'])
        parser.add_option("-l", "--log",
            help="Filename for logging (by default %s)" %(repr(self.DEFAULT_OPTIONS['log'])),
            action="store", type="string", metavar="LOCATION",
            default=self.DEFAULT_OPTIONS['log'])
        parser.add_option("-p", "--pidfile",
            help="Filename for pid (by default %s)" %(repr(self.DEFAULT_OPTIONS['pidfile'])),
            action="store", type="string", metavar="LOCATION",
            default=self.DEFAULT_OPTIONS['pidfile'])

        # Other Options
        self.additionalOptions(parser)

        # Parse options
        options, arguments = parser.parse_args()

        # Check option values
        if options.verbosity not in range(len(self.LEVELS)):
            print >>sys.stderr, "Usage error: Verbosity level must be between 0 and %s."  %(len(self.LEVELS)-1)
            sys.exit(1)
        if hasattr(options, 'not_daemonize'):
            options.daemonize = not options.not_daemonize
            delattr(options, 'not_daemonize')
        
        # Check other option values
        self.additionalOptionCheck(options, arguments)
        
        self.options = options
        self.arguments = arguments
                
    def additionalOptions(self, parser):
        # Override this method if needed
        pass
        # Example of option group
        #optgrp = OptionGroup(parser, "Filter")
        #optgrp.add_option("-s", "--orig-src", help="Source address from original direction",
        #    action="store", type="string", default=None)
        #parser.add_option_group(optgrp)

    def additionalOptionCheck(self, options, arguments):
        # Override this method if needed
        pass
        # Example of option check
        #if values.orig_src:
        #    values.orig_src = IP(values.orig_src)
    
    def __startLogging(self):
        #TODO: enabling other logging methods (by using other handler types)
        log = logging.getLogger(self.APP_NAME)
        logfile = logging.FileHandler(self.options.log)
        formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] %(message)s')
        logfile.setFormatter(formatter)
        log.addHandler(logfile)  
        log.setLevel(self.LEVELS[self.options.verbosity])
        # Logging itself ;)
        log.info("%(program_name)s started with verbosity %(verbosity_level)s (PID: %(pid)s)" \
                                %{'program_name': self.APP_ABSPATH,\
                                'verbosity_level': self.options.verbosity,
                                'pid': os.getpid()})
        
    def __stopLogging(self):
        log = logging.getLogger(self.APP_NAME)
        log.info("%(program_name)s stopped" \
                                %{'program_name': self.APP_ABSPATH})
        logging.shutdown()
        
    def __launchToBackground(self):
        # Source: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66012/index_txt
        # Author: Jürgen Hermann
        # do the UNIX double-fork magic, see Stevens' "Advanced 
        # Programming in the UNIX Environment" for details (ISBN 0201563177)
        pidfilename = self.options.pidfile
        if os.path.exists(pidfilename):
            print >>sys.stderr, "PIDFile error: PIDfile already exists: %s" %(pidfilename) 
            sys.exit(1)
        try:
            pidfile = open(pidfilename, 'w')
        except IOError, e:
            print >>sys.stderr, "IO error: Unable to open PIDfile: %d (%s)" % (e.errno, e.strerror) 
            sys.exit(1)
        except OSError, e:
            print >>sys.stderr, "IO error: Unable to open PIDfile: %d (%s)" % (e.errno, e.strerror) 
            sys.exit(1)
        try: 
            pid = os.fork()
            if pid > 0:
                # Exit first parent, sleep to allow PID print from 2nd fork
                time.sleep(1)
                sys.exit(0) 
        except OSError, e: 
            print >>sys.stderr, "Fork error: fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
            sys.exit(1)
        # Decouple from parent environment
        os.chdir("/") 
        os.setsid() 
        os.umask(0)
        # do second fork
        try: 
            pid = os.fork() 
            if pid > 0:
                # Exit from second parent, print eventual PID before
                print "daemon PID %d" % pid 
                sys.exit(0) 
        except OSError, e: 
            print >>sys.stderr, "Fork error: fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
            sys.exit(1)
        pidfile.write(str(os.getpid()))
        pidfile.close()

    def __deletePidfile(self):
        try:
            os.remove(self.options.pidfile)
        except IOError, e:
            log = logging.getLogger(self.APP_NAME)
            log.warning("Unable to remove the pidfile : %d (%s)" %(e.errno, e.strerror))
        except OSError, e:
            log = logging.getLogger(self.APP_NAME)
            log.warning("Unable to remove the pidfile : %d (%s)" %(e.errno, e.strerror))

    def start(self):
        if self.options.daemonize:
            self.__launchToBackground()
        self.__startLogging()
        self.oldsig = (signal.getsignal(signal.SIGTERM), signal.getsignal(signal.SIGINT))
        signal.signal(signal.SIGTERM, self.interrupt)
        signal.signal(signal.SIGINT, self.interrupt)
        # Call subclass' run method
        try:
            self.run()
        except SystemExit:
            pass
        except:
            log = logging.getLogger(self.APP_NAME)
            log.exception("Unhandled exception while daemon is running")
            self.stop()
            sys.exit(2)

    def stop(self):
        if self.options.daemonize:
            try:
                self.__deletePidfile()
            except:
                log = logging.getLogger(self.APP_NAME)
                log.exception("Unhandled exception while is exiting")
        self.__stopLogging()

    def interrupt(self, sig, frame):
        signal.signal(signal.SIGTERM, self.oldsig[0])
        signal.signal(signal.SIGINT, self.oldsig[1])
        self.stop()
        sys.exit(0)

    def run(self):
        # Override this method
        pass

if __name__ == "__main__":
    # This is a example of usage
    def myrun(): time.sleep(3)
    daemon = AutoDaemon("test_AutoDaemon")
    daemon.run = myrun # To emulate subclass overriding
    daemon.parseOptions()
    daemon.start()
    
    
    
    
