#!/usr/bin/env python
# -*- coding: utf-8 -*-

from SimpleXMLRPCServer import SimpleXMLRPCServer
import logging, thread, time, random, sha
from socket import ntohs
from auto_daemon import AutoDaemon
from pynetfilter_conntrack import ATTRIBUTES
from viewer import ConntrackViewer

"""
Copyright(C) 2006, 2007 INL
Written by Damien Boucard <damien.boucard AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
---
server.py is a part of pyctd which manages XML-RPC service.
"""
class PyctServer(SimpleXMLRPCServer):
    """
    PyctServer is a Simple XMLRPC Server which serves Pyctd queries
    (from Conntrack PHP interface, for instance).
    
    Instance attributes:
     - log (Logger): object to write log messages
     - data: DataCollector instance
     - authenticate (func): function from the ``auth`` plugin passed in
                            command-line argument which is run to
                            authenticate XMLRPC clients
    
    Class attributes:
     - allow_reuse_address (bool): True to be able to use a port which
                                   has not been cleanly closed
     - SESSION_DURATION (int): duration before the web session expires
    """
    allow_reuse_address = True
    SESSION_DURATION = 15 # seconds
    
    def __init__(self, data, host, port, mod_auth):
        self.log = logging.getLogger(AutoDaemon.APP_NAME)
        self.data = data
        if mod_auth is not None:
            self.authenticate = mod_auth.authenticate
        else:
            self.authenticate = self.__authenticate
        self.__sessions = {}
        self.__session_lock = thread.allocate_lock()
        SimpleXMLRPCServer.__init__(self, (host, port), logRequests = False)

    def _dispatch(self, method, params):
        """
        Dispatch remote procedures. All remote procedures will produce the
        calling the method named as the procedure, prefixed by "rpc_"
        (excepted for ``login`` procedure).
        
        Params are passed as argument of these methods.
        
        """
        self.update_sessions()
        if method == "login":
            if self.authenticate(*params):
                result = self.create_session(*params)
                self.log.info("'%s' is successfully authenticated" %(params[0]))
            else:
                self.log.error("Authentication of %s failed" %(params[0]))
        else:
            params = list(params)
            id = params.pop(0)
            if id not in self.__sessions:
                raise Exception('Wrong id or session has timed out.')
            try:
                # We are forcing the 'rpc_' prefix on methods that are
                # callable through XML-RPC to prevent potential security
                # problems
                func = getattr(self, 'rpc_' + method)
            except AttributeError:
                self.log.error('Method "%s" is not supported' % method)
                raise Exception('Method "%s" is not supported' % method)
            try:
                result = func(*params)
            except:
                self.log.exception("Unhandled exception while serving %s%s" %(func.__name__, tuple(params)))
                raise Exception('Internal server error')
            self.log.info('Method "%s%s" is successfully called' %(method, tuple(params)))
        return result

    def __authenticate(self, *params):
        """
        Default authentication function if no plugin has been passed in the
        ``--auth=`` option of the command-line.
        
        This dummy function returns always True.
        """
        return True

    def create_session(self, username, password):
        """
        Create a web session for XMLRPC clients.
        
        The session is identified by ID which must be given in 1st
        argument of each procedure calls (excepted ``login``).
        
        The session automatically expires self.SESSION_DURATION seconds
        after the session creation.
        """
        self.__session_lock.acquire()
        try:
            id = sha.new(str(random.random())).hexdigest()
            while id in self.__sessions:
                id = sha.new(str(random.random())).hexdigest()
            self.__sessions[id] = time.time() + self.SESSION_DURATION
        finally:
            self.__session_lock.release()
        return id
    
    def update_sessions(self):
        """
        Check for session expiration and delete it if it is the case.
        """
        t = time.time()
        self.__session_lock.acquire()
        try:
            for (id, timeout) in self.__sessions.copy().items():
                if timeout < t:
                    del self.__sessions[id]
        finally:
            self.__session_lock.release()
    
    def rpc_view(self, filters, order_by, order_dir, start, length):
        """
        ``view`` remote procedure handler.
        
        Parameters:
         - filters (dict): field name (str): field value (type depends to the field)
         - order_by (str): field name ordered by
         - order_dir (str): order direction, 'A' for ascending and 'D' for descending
         - start (int): first connection index to render (used for pagination)
         - length (int): maximal number of connections to render (used for pagination)
        
        Return a dict containing a list of connections formatted to be
        XMLRPC compliant.
        """
        viewer = ConntrackViewer()
        viewer.set_order(order_by, order_dir)
        viewer.set_filters(filters)
        viewer.set_pagination(start, length)
        return viewer.xmlrpc_view(self.data.current)
    
    def rpc_kill(self, conn_ids):
        """
        ``kill`` remote procedure handler.
        
        Parameters:
         - conn_ids (list): connection identifiers to destroy (int)
        
        Return 1 if connections killed successfully, else 0.
        """
        return (self.data.kill(conn_ids) and 0 or 1)
    
    def rpc_change(self, id, field, value):
        """
        ``change`` remote procedure handler.
        
        Parameters:
         - id (int): connection identifier to modify
         - field (str): field name of the connection to change
         - value (int): new value to set to the field
        
        Note: the type of value parameter is int because only changeable fields
        are integer-typed for the moment, but it can change in future versions.
        """
        try:
            return (self.data.change(id, field, value) and 0 or 1)
        except (TypeError, ValueError):
            return 2

    def rpc_refresh(self):
        """
        ``refresh`` remote procedure handler.

        Return 0.
        """
        self.data.refresh()
        return 0

    def rpc_set_autorefresh(self, autorefresh):
        """
        ``toggle_autorefresh`` remote procedure handler.
        Return 0.
        """
        self.data.setRefreshing(autorefresh)
        return 0


