#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright(C) 2006,2007 INL
Written by Damien Boucard <damien.boucard AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
---
viewer.py is a part of pyctd which generates the XML-RPC response for the
"view" procedure.
"""
from xmlrpclib import DateTime
import pynetfilter_conntrack as nfct
import IPy

ASCENDING_SORT = 'A'
DESCENDING_SORT = 'D'

class ConntrackField:
    """
    Data structure for managing conntrack field which can be displayed
    to XMLRPC clients.
    
    Attributes:
     - name (str): name of the field (identifier)
     - type (str): type of the field value
     - orderable (bool): if we can order the connection set by this field
     - filterable (bool): if we can filter the connection set on this field
     - nfct_field (str): related name of the field in the 
                         libnetfilter_conntrack API
    """
    def __init__(self, name, type="int", nfct_field=None, filterable=True, orderable=True):
        self.name = name
        self.type = type
        self.orderable = orderable
        self.filterable = filterable
        self.nfct_field = nfct_field

fields = {
    "orig_ipv4_src": ConntrackField("orig_ipv4_src", "ip", "orig_ipv4_src"),
    "orig_ipv4_dst": ConntrackField("orig_ipv4_dst", "ip", "orig_ipv4_dst"),
    "repl_ipv4_src": ConntrackField("repl_ipv4_src", "ip", "repl_ipv4_src"),
    "repl_ipv4_dst": ConntrackField("repl_ipv4_dst", "ip", "repl_ipv4_dst"),
    "orig_ipv6_src": ConntrackField("orig_ipv6_src", "ip", "orig_ipv6_src"),
    "orig_ipv6_dst": ConntrackField("orig_ipv6_dst", "ip", "orig_ipv6_dst"),
    "repl_ipv6_src": ConntrackField("repl_ipv6_src", "ip", "repl_ipv6_src"),
    "repl_ipv6_dst": ConntrackField("repl_ipv6_dst", "ip", "repl_ipv6_dst"),
    "orig_port_src": ConntrackField("orig_port_src", "int", "orig_port_src"),
    "orig_port_dst": ConntrackField("orig_port_dst", "int", "orig_port_dst"),
    "repl_port_src": ConntrackField("repl_port_src", "int", "repl_port_src"),
    "repl_port_dst": ConntrackField("repl_port_dst", "int", "repl_port_dst"),
    "icmp_type": ConntrackField("icmp_type", "int", "icmp_type"),
    "icmp_code": ConntrackField("icmp_code", "int", "icmp_code"),
    "icmp_id": ConntrackField("icmp_id", "int", "icmp_id"),
    "orig_l3proto": ConntrackField("orig_l3proto", "int", "orig_l3proto"),
    "repl_l3proto": ConntrackField("repl_l3proto", "int", "repl_l3proto"),
    "orig_l4proto": ConntrackField("orig_l4proto", "int", "orig_l4proto"),
    "repl_l4proto": ConntrackField("repl_l4proto", "int", "repl_l4proto"),
    "tcp_state": ConntrackField("tcp_state", "int", "tcp_state"),
    "snat_ipv4": ConntrackField("snat_ipv4", "ip", "snat_ipv4"),
    "dnat_ipv4": ConntrackField("dnat_ipv4", "ip", "dnat_ipv4"),
    "snat_port": ConntrackField("snat_port", "int", "snat_port"),
    "dnat_port": ConntrackField("dnat_port", "int", "dnat_port"),
    "timeout": ConntrackField("timeout", "int", "timeout"),
    "mark": ConntrackField("mark", "int", "mark"),
    "orig_counter_packets": ConntrackField("orig_counter_packets", "int", "orig_counter_packets", False),
    "repl_counter_packets": ConntrackField("repl_counter_packets", "int", "repl_counter_packets", False),
    "orig_counter_bytes": ConntrackField("orig_counter_bytes", "int", "orig_counter_bytes", False),
    "repl_counter_bytes": ConntrackField("repl_counter_bytes", "int", "repl_counter_bytes", False),
    "orig_byterate": ConntrackField("orig_byterate", "float", filterable=False),
    "repl_byterate": ConntrackField("repl_byterate", "float", filterable=False),
    "use": ConntrackField("use", "int", "use"),
    "id": ConntrackField("id", "int"),
    "status": ConntrackField("status", "int", "status", False, False),
    "username": ConntrackField("username", "string"),
    "authenticated": ConntrackField("authenticated", "boolean"),
}

class ConntrackViewer:
    """
    This class is used to convert a conntrack table in a XMLRPC-compliant
    format in order to send it to clients.
    
    Instance attributes:
     - filters (dict): field name (str): field value (type depends to the field)
     - order_by (str): field name ordered by
     - order_dir (str): order direction, ASCENDING_SORT or DESCENDING_SORT
     - start (int): first item to render (used for pagination)
     - length (int): maximal number of connections to render (used for pagination)
    
    Class attributes:
     - DEFAULT_ORDER_FIELD (str): default order field if the order_by
                                  attribute is not set
     - DEFAULT_START (int): default item start value if the start
                            attribute is not set
     - DEFAULT_LENGTH (int): default number of items if the length
                             attribute is not set
    """
    DEFAULT_ORDER_FIELD = 'id'
    DEFAULT_START = 0
    DEFAULT_LENGTH = None
    
    def __init__(self):
        self.order_by = self.DEFAULT_ORDER_FIELD
        self.order_dir = DESCENDING_SORT
        self.filters = {}
        self.start = self.DEFAULT_START
        self.length = self.DEFAULT_LENGTH
    
    def set_order(self, order_by, order_dir):
        if order_by not in fields:
            raise ValueError('Unknown field name: "%s"' %(order_by))
        if not fields[order_by].orderable:
            raise ValueError('The field named "%s" cannot be sorted.' %(order_by))
        if order_dir not in (DESCENDING_SORT, ASCENDING_SORT):
            raise ValueError('Invalid direction value: "%s"; Must be "%s" or "%s".'
                                 %(order_dir, ASCENDING_SORT, DESCENDING_SORT))
        self.order_by = order_by
        self.order_dir = order_dir
    
    def set_filters(self, filter_dict):
        filters = {}
        for name, value in filter_dict.iteritems():
            if name not in fields:
                raise KeyError('Unknown field name: "%s"' %(name))
            field = fields[name]
            if not field.filterable:
                raise ValueError('The field named "%s" cannot be filtered.' %(name))
            if field.type == "ip":
                try:
                    ip = IPy.IP(value)
                except ValueError:
                    raise ValueError('%s: Invalid IP address or network: "%s"'
                                                                %(name, value))
                value = ip
            elif name.find("port") != -1:
                if isinstance(value, str):
                    try: value = int(value)
                    except ValueError: pass
                if not isinstance(value, int) or not 0 < value < 65536:
                    raise ValueError('%s: Invalid port number: "%s"'
                                                                %(name, value))
            elif field.type in ("int","boolean"):
                try:
                    integer = int(value)
                except ValueError:
                    raise ValueError('%s: Invalid field value: "%s"; string representing an integer expected'
                                                                %(name, value))
                value = integer
            filters[name] = value
        self.filters = filters
    
    def set_pagination(self, start, length=None):
        if type(start) is not int:
            raise ValueError('"start" argument must be an integer; %s (%s) given.'
                                                         %(start, type(start)))
        if length is not None and type(length) is not int:
            raise ValueError('"length" argument must be an integer; %s (%s) given.'
                                                       %(length, type(length)))
        self.start = start
        self.length = length
    
    def xmlrpc_view(self, conntrack_table):
        """
        conntrack_table is an instance of ConnectionSet.
        
        The method returns a dict representing the conntrack table in
        an XMLRPC-compliant data structure.
        """
        connset = conntrack_table
        root = {}
        root["timestamp"] = DateTime(connset.hit)
        root["have_usernames"] = int(connset.uid_mask is not None)
        root["total"] = len(connset)
        nb_filtered = 0
        conntab = []
        for id, conn in connset.iteritems():
            # Apply filters
            keep = True
            for name, fvalue in self.filters.iteritems():
                if fields[name].nfct_field == None:
                    value = getattr(conn, name)
                else:
                    value = getattr(conn.conntrack, name)
                is_network = False
                if (fields[name].type == "ip" and IPy.IP(value) not in fvalue) or \
                   (fields[name].type != "ip" and value != fvalue):
                    keep = False
                    break
            if not keep:
                continue
            nb_filtered += 1
            # End of filtering
            
            # Serialize one connection #
            conntrack = {}
            for field in fields.itervalues():
                if field.nfct_field is None:
                    value = str(getattr(conn, field.name))
                else:
                    value = getattr(conn.conntrack, field.nfct_field)
                    if  nfct.ATTRIBUTES[field.nfct_field][1] >= 32:
                        # Note: XML-RPC does not manage values over 32 bits
                        value = str(value)
                if field.type == "boolean":
                    value = int(value == 'True')
                conntrack[field.name] = value
            conntab.append(conntrack)
            # End of connection of serialization #
        conntab.sort(self.sort_connections)
        if self.length is None:
            conntab = conntab[self.start:]
        else:
            conntab = conntab[self.start:self.start+self.length]
        root["table"] = conntab
        root["filtered"] = nb_filtered
        return root
    
    def sort_connections(self, conn1, conn2):
        if self.order_dir == DESCENDING_SORT:
            conn1, conn2 = conn2, conn1
        if fields[self.order_by].type == "int":
            val1, val2 = int(conn1[self.order_by]), int(conn2[self.order_by])
        elif fields[self.order_by].type == "float":
            val1, val2 = float(conn1[self.order_by]), float(conn2[self.order_by])
        else:
            val1, val2 = conn1[self.order_by], conn2[self.order_by]
        #print "cmp(%r, %r) <==> %r" %(conn1[self.order_by], conn2[self.order_by], cmp(conn1[self.order_by], conn2[self.order_by]))
        return cmp(val1, val2)